# Copyright 2025 Cornell University
# released under BSD 3-Clause License
# author: Kevin Laeufer <laeufer@cornell.edu>

import pathlib
import pytest
from pypatronus import *

repo_root = (pathlib.Path(__file__) / '..' / '..' / '..').resolve()



def test_transition_system_simulation():
    sys = parse_btor2_file(repo_root / "inputs" / "unittest" / "swap.btor")
    sim = Interpreter(sys)
    a, b = sys['a'].symbol, sys['b'].symbol

    sim.init()
    assert sim[a] == 0, "a@0"
    assert sim[b] == 1, "b@0"

    sim.step()
    assert sim[a] == 1
    assert sim[b] == 0

    sim.step()
    assert sim[a] == 0
    assert sim[b] == 1



# Copyright 2025 Cornell University
# released under BSD 3-Clause License
# author: Kevin Laeufer <laeufer@cornell.edu>

import pathlib
import pytest
from pypatronus import *

repo_root = (pathlib.Path(__file__) / '..' / '..' / '..').resolve()

def test_call_smt_solver():
    a = BitVec('a', 3)
    b = BitVec('b', 3)
    s = Solver('z3')
    r = s.check(a < b)
    assert str(r) == "sat"

    r = s.check(a < b, a > b)
    assert str(r) == "unsat"

    # to generate a model, we need to actually add the assertion!
    s.add(a < b)
    s.check()
    m = s.model()
    assert len(m) == 2
    assert isinstance(m[a], int)
    assert isinstance(m[b], int)
    assert m[a] < m[b]


def test_parse_smt_lib_expr():
    symbols = {
        'x_0': BitVec('x_0', 32),
        'y_0': BitVec('y_0', 32),
        'x_1': BitVec('x_1', 32),
    }
    a = parse_smtlib_expr("(= x_1 (bvadd x_0 y_0))", symbols)
    assert str(a) == "eq(x_1, add(x_0, y_0))"


@pytest.mark.skip(reason="parsing commands is not implemented yet")
def test_parse_smt_lib_commands():
    parse_smtlib_cmd("(set-logic QF_BV)")
    parse_smtlib_cmd("(set-option :produce-models true)")
    parse_smtlib_cmd("(declare-const x_0 (_ BitVec 32))")



# Copyright 2025 Cornell University
# released under BSD 3-Clause License
# author: Kevin Laeufer <laeufer@cornell.edu>

from pypatronus import *

def test_simplify():
    # by default this uses a global simplifier
    true = BitVecVal(1, 1)
    false = BitVecVal(0,1)
    a = BitVec('a', 1)
    assert simplify((~a) & a) == false
    assert simplify((~a) | a) == true

    assert simplify(SignExt(1, false)) == BitVecVal(0b00, 2)
    assert simplify(SignExt(1, true)) == BitVecVal(0b11, 2)

    assert simplify(BitVecVal(0, 4).equals(Extract(8, 5, ZeroExt(4, BitVec('a', 5))))) == true





# Copyright 2025 Cornell University
# Copyright 2025 The Regents of the University of California
# released under BSD 3-Clause License
# author: Kevin Laeufer <laeufer@cornell.edu>
# author: Adwait Godbole <adwait@berkeley.edu>

import pathlib
import pytest
from pypatronus import *


repo_root = (pathlib.Path(__file__) / '..' / '..' / '..').resolve()

COUNT_2 = """
1 sort bitvec 3
2 zero 1
3 state 1
4 init 1 3 2
5 one 1
6 add 1 3 5
7 next 1 3 6
8 ones 1
9 sort bitvec 1
10 eq 9 3 8
11 bad 10
"""

def test_parse_and_serialize_count2():
    sys = parse_btor2_str(COUNT_2, "count2")
    assert sys.name == "count2"

    expected_system = """
count2
bad _bad_0 : bv<1> = eq(_state_0, 3'b111)
state _state_0 : bv<3>
  [init] 3'b000
  [next] add(_state_0, 3'b001)
    """
    assert expected_system.strip() == str(sys).strip()

@pytest.mark.skip(reason="btor2 serialization is not yet implemented in paronus")
def btor2_serialize():
    sys = parse_btor2_str(COUNT_2, "count2")
    assert sys.to_btor2_str().strip() == COUNT_2.strip()

def test_transition_system_fields():
    sys = parse_btor2_str(COUNT_2, "count2")
    assert sys.inputs == []
    assert sys.constraints == []
    assert [str(e) for e in sys.bad_states] == ["eq(_state_0, 3'b111)"]
    assert len(sys.states) == 1
    state = sys.states[0]
    assert state.name == "_state_0"
    assert str(state.symbol) == "_state_0"
    assert str(state.init) == "3'b000"
    assert str(state.next) == "add(_state_0, 3'b001)"
    assert len(sys.outputs) == 0


def test_expression_builder():
    # we are emulating the Z3 API as much as possible
    a = BitVec('a', 3)
    b = BitVec('b', 3)
    assert str(a < b) == "sgt(b, a)"


def test_transition_system_builder():
    sys = TransitionSystem("test")
    en, count_s = BitVec('en', 1), BitVec('count_s', 8)
    sys.inputs = [en]
    sys.states = [State('count_s', init=BitVecVal(0, 8), next=If(en, count_s + BitVecVal(1, 8), count_s))]
    sys.outputs = [Output('count', count_s)]
    # TODO: there is a big pitfall here: you cannot just `append` to the bad_states, inputs, etc. because we use
    #       a getter / setter approach
    sys.add_bad_state('count_is_123', count_s.equals(BitVecVal(123, 8)))
    expected_system = """
test
input en : bv<1>
output count : bv<8> = count_s
bad count_is_123 : bv<1> = eq(count, 8'b01111011)
state count_s : bv<8>
  [init] 8'b00000000
  [next] ite(en, add(count, 8'b00000001), count)
    """
    assert str(sys).strip() == expected_system.strip()





# Copyright 2025 Cornell University
# released under BSD 3-Clause License
# author: Kevin Laeufer <laeufer@cornell.edu>

import pathlib
import pytest
from pypatronus import *

repo_root = (pathlib.Path(__file__) / '..' / '..' / '..').resolve()



def test_transition_system_model_checking():
    sys = parse_btor2_file(repo_root / "inputs" / "unittest" / "swap.btor")
    mc = SmtModelChecker('z3')
    # there are no assertions in this circuit, so it cannot fail
    assert str(mc.check(sys, 4)) == "unsat"

    # add an assertion
    a = next(s.symbol for s in sys.states if str(s.symbol) == 'a')
    b = next(s.symbol for s in sys.states if str(s.symbol) == 'b')
    sys.add_assertion("a_is_0", a.equals(BitVecVal(0, 8)))
    r = mc.check(sys, 4)
    assert str(r) == "sat"

    # check model
    assert len(r) == 2, "fail in step 2"

    # check initial values
    # TODO!
    # assert r[a] == BitVecVal(0, 8)
    # assert r[b] == BitVecVal(1, 8)